import matplotlib.pyplot as plt
from sklearn.metrics import ConfusionMatrixDisplay

from trava.metric import Metric


def plot_confusion_matrix(metric: Metric, fig, ax, color: str, label: str):
    ax.set_title(label + ' ' + 'Confusion matrix')
    conf_matrix = metric.value
    class_names = list(range(len(conf_matrix)))
    ConfusionMatrixDisplay(conf_matrix, class_names).plot(cmap=plt.cm.Blues, ax=ax)



import pickle
from trava.model_serializer import ModelSerializer


class PickleModelSerializer(ModelSerializer):
    def load(self, path: str):
        with open(path, 'r') as file:
            return pickle.load(file)

    def save(self, model, path: str):
        with open(path, 'wb') as file:
            return pickle.dump(model, file)



import pytest
import numbers
import numpy as np
import pandas as pd

from tests.objects_for_tests import TestResultsHandler, TestScorer, TestAnyScorer, TestModel
from trava.evaluator import Evaluator
from trava.fit_predictor import FitPredictConfig
from trava.metric import Metric
from trava.raw_dataset import RawDataset
from trava.split.result import SplitResult
from trava.trava_tracker import TravaTracker
from trava.trava_sv import TravaSV


@pytest.fixture(scope='class')
def trava_tracker():
    return TravaTracker(scorers=[])


@pytest.fixture(scope='class')
def test_result_handlers():
    return [TestResultsHandler(scorers=[TestScorer(score_func=np.sum)])]


@pytest.fixture
def trava(test_result_handlers, trava_tracker):
    return TravaSV(results_handlers=test_result_handlers, tracker=trava_tracker)


@pytest.fixture(scope='class')
def split_result(X_y_train, X_y_test):
    X_train, y_train = X_y_train
    X_test, y_test = X_y_test

    return SplitResult(X_train=X_train,
                       y_train=y_train,
                       X_test=X_test,
                       y_test=y_test)


@pytest.fixture(scope='class')
def X_y_train():
    X_train = pd.DataFrame({
        'f1': [1, 2, 3],
    })
    y_train = np.array([0, 0, 0])

    return X_train, y_train


@pytest.fixture(scope='class')
def X_y_test():
    X_test = pd.DataFrame({
        'f1': [4, 5, 6]
    })
    y_test = np.array([1, 1, 1])

    return X_test, y_test


@pytest.fixture(scope='class')
def raw_dataset():
    result = RawDataset(df=pd.DataFrame({'df': [1, 2, 3], 'test_target': [0, 0, 1]}), target_col_name='test_target')
    return result


@pytest.fixture(scope='class')
def model_init_params():
    model_init_params = {
        'test1': 1,
        'test2': 'second',
    }
    return model_init_params


@pytest.fixture(scope='class')
def model_id():
    return 'test_model_id'


@pytest.fixture(scope='class')
def main_model_id(model_id):
    return 'main_' + model_id


@pytest.fixture()
def raw_model(mocker):
    model_mock = mocker.Mock()
    return model_mock


@pytest.fixture()
def trava_model(mocker, raw_model):
    model_mock = mocker.Mock()
    trava_model.raw_model = raw_model
    return model_mock


@pytest.fixture
def model_type(mocker, raw_model):
    model_type = mocker.stub()
    model_type.return_value = raw_model

    return model_type


def _fit_predictor(mocker, evaluators):
    fit_predictor = mocker.Mock()
    fit_predictor.fit_predict.return_value = evaluators
    return fit_predictor


def _mock_evaluator(mocker,
                    model_id,
                    train_values=None,
                    test_values=None,
                    any_values=None,
                    trava_model=None,
                    split_result=None):
    evaluator = mocker.Mock()
    evaluator.model_id = model_id

    if train_values:
        train_metrics = [_metric(value=value, model_id=model_id) for value in train_values]
        evaluator.train_metrics.return_value = train_metrics

    if test_values:
        test_metrics = [_metric(value=value, model_id=model_id) for value in test_values]
        evaluator.test_metrics.return_value = test_metrics

    if any_values:
        any_metrics = [_metric(value=value, model_id=model_id) for value in any_values]
        evaluator.other_metrics.return_value = any_metrics

    if trava_model:
        evaluator.trava_model = trava_model

    if split_result:
        evaluator.fit_split_data = split_result

    if raw_dataset:
        evaluator.raw_split_data = split_result

    return evaluator


def _true_evaluator(model,
                    split_result):
    return Evaluator(trava_model=model, fit_split_data=split_result, raw_split_data=split_result)


def _metric(model_id, value):
    return Metric(name=model_id + '_' + str(value), value=value, model_id=model_id)


def _fit_with_n_evaluators(mocker,
                           trava_model,
                           model_id,
                           main_model_id,
                           model_type,
                           split_result,
                           trava,
                           n_evaluators,
                           test_result_handlers,
                           keep_data_in_memory: bool = True,
                           mock_evaluators: bool = True):
    evaluators = []
    for i in range(n_evaluators):
        indexed_model_id = model_id + '_' + str(i)
        eval_model_id = indexed_model_id if n_evaluators > 1 else model_id
        if mock_evaluators:
            evaluator = _mock_evaluator(mocker,
                                        model_id=eval_model_id,
                                        train_values=mocker.MagicMock(),
                                        test_values=mocker.MagicMock(),
                                        any_values=mocker.MagicMock(),
                                        trava_model=trava_model,
                                        split_result=split_result)
        else:
            trava_model.model_id = eval_model_id
            evaluator = _true_evaluator(model=trava_model, split_result=split_result)
            # since fit predictor is mocked I need to evaluate the true evaluator manually
            evaluator.evaluate(scorers_providers=test_result_handlers)
        evaluators.append(evaluator)

    fit_predictor = _fit_predictor(mocker=mocker, evaluators=evaluators)
    trava.fit_predict(raw_split_data=split_result,
                      model_id=main_model_id if n_evaluators > 1 else model_id,
                      model_type=model_type,
                      fit_predictor=fit_predictor,
                      keep_data_in_memory=keep_data_in_memory)
    return evaluators


# TESTS

def test_model_init(trava: TravaSV,
                    split_result: SplitResult,
                    model_init_params,
                    model_type,
                    model_id):
    trava.fit_predict(raw_split_data=split_result,
                      model_id=model_id,
                      model_type=model_type,
                      model_init_params=model_init_params)

    model_type.assert_called_once_with(**model_init_params)


def test_fit_predictor(mocker,
                       raw_model,
                       trava: TravaSV,
                       raw_dataset: RawDataset,
                       split_result: SplitResult,
                       model_init_params,
                       model_type,
                       model_id,
                       test_result_handlers,
                       trava_tracker):
    fit_predictor = _fit_predictor(mocker=mocker, evaluators=[])

    keep_models_in_memory = True
    serializer = mocker.Mock()
    fit_params = {'a': 1, 'b': 2}
    predict_params = {'c': 3, 'd': 4}
    description = 'test_description'
    trava.fit_predict(raw_split_data=split_result,
                      raw_dataset=raw_dataset,
                      model_id=model_id,
                      model_type=model_type,
                      description=description,
                      model_init_params=model_init_params,
                      fit_predictor=fit_predictor,
                      fit_params=fit_params,
                      predict_params=predict_params,
                      keep_models_in_memory=keep_models_in_memory,
                      serializer=serializer)

    fit_predictor.fit_predict.assert_called_once()

    test_fit_config = FitPredictConfig(raw_split_data=split_result,
                                       raw_dataset=raw_dataset,
                                       raw_model=raw_model,
                                       model_init_params=model_init_params,
                                       model_id=model_id,
                                       scorers_providers=test_result_handlers + [trava_tracker],
                                       serializer=serializer,
                                       fit_params=fit_params,
                                       predict_params=predict_params,
                                       description=description)

    fit_predictor.fit_predict.assert_called_once_with(config=test_fit_config, tracker=trava_tracker)


@pytest.mark.parametrize("evaluator_values", [
    [([1.0, 3.0, {}, 'train_string'],  # train
      [4.0, 4.0, [], 'test_string'],   # test
      [10, pd.DataFrame, []])],        # model meta

    [([1.0, 3.0, {}, 'train_string'],  # train
      [4.0, 4.0, [], 'test_string'],   # test
      [10, pd.DataFrame, []]),         # model meta

     ([9.0, 5.0, {}, 'train_string'],  # train
      [16.0, 8.0, [], 'test_string'],  # test
      [20, pd.DataFrame, []])          # model meta
     ]
])
def test_fit_predict_many_result(mocker,
                                 trava: TravaSV,
                                 split_result: SplitResult,
                                 model_type,
                                 model_id,
                                 evaluator_values):
    evaluators = []

    for values in evaluator_values:
        evaluator = _mock_evaluator(mocker,
                                    model_id=model_id,
                                    train_values=values[0],
                                    test_values=values[1],
                                    any_values=values[2])

        evaluators.append(evaluator)

    fit_predictor = _fit_predictor(mocker=mocker, evaluators=evaluators)

    result = trava.fit_predict(raw_split_data=split_result,
                               model_id=model_id,
                               model_type=model_type,
                               fit_predictor=fit_predictor)
    if len(evaluator_values) == 1:
        assert result == [[evaluator_values[0][0], evaluator_values[0][1], evaluator_values[0][2]]]
    else:
        assert result == [[[5.0, 4.0], [10.0, 6.0], [15.0]]]

    assert trava.detailed_results_for(model_id=model_id)


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_no_save(mocker,
                          trava,
                          model_id,
                          main_model_id,
                          model_type,
                          trava_model,
                          split_result,
                          test_result_handlers,
                          n_evaluators):
    evaluators = _fit_with_n_evaluators(mocker=mocker,
                                        trava_model=trava_model,
                                        model_id=model_id,
                                        main_model_id=main_model_id,
                                        model_type=model_type,
                                        split_result=split_result,
                                        trava=trava,
                                        n_evaluators=n_evaluators,
                                        test_result_handlers=test_result_handlers)

    trava.evaluate(model_id=main_model_id if n_evaluators > 1 else model_id,
                   results_handlers=test_result_handlers,
                   save_results=False)

    for evaluator in evaluators:
        evaluator.evaluate.assert_not_called()


# noinspection PyPep8Naming
@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_save(mocker,
                       trava,
                       model_id,
                       main_model_id,
                       raw_model,
                       trava_model,
                       model_type,
                       split_result,
                       X_y_train,
                       X_y_test,
                       n_evaluators):
    any_and_test_data = 1233
    trava_model.get_model.return_value = raw_model
    trava_model.any_and_test_data = any_and_test_data

    result_handlers = [TestResultsHandler(scorers=[
        TestScorer(score_func=np.sum),
        TestAnyScorer(model_func=lambda scorer_model: scorer_model.any_and_test_data)
    ])]

    evaluators = _fit_with_n_evaluators(mocker=mocker,
                                        trava_model=trava_model,
                                        model_id=model_id,
                                        main_model_id=main_model_id,
                                        model_type=model_type,
                                        split_result=split_result,
                                        trava=trava,
                                        n_evaluators=n_evaluators,
                                        test_result_handlers=result_handlers)

    result = trava.evaluate(model_id=main_model_id if n_evaluators > 1 else model_id,
                            results_handlers=result_handlers,
                            save_results=True)

    for evaluator in evaluators:
        evaluator.evaluate.assert_called_once_with(scorers_providers=result_handlers)

    X_train, _ = X_y_train
    X_test, _ = X_y_test

    f1_train_sum = np.sum(X_train['f1'])
    f1_test_sum = np.sum(X_test['f1'])

    assert result == [[[f1_train_sum], [f1_test_sum], [any_and_test_data]]]


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_track(mocker,
                        trava,
                        model_id,
                        main_model_id,
                        model_type,
                        trava_model,
                        split_result,
                        test_result_handlers,
                        X_y_train,
                        X_y_test,
                        n_evaluators):
    evaluators = _fit_with_n_evaluators(mocker=mocker,
                                        trava_model=trava_model,
                                        model_id=model_id,
                                        main_model_id=main_model_id,
                                        model_type=model_type,
                                        split_result=split_result,
                                        trava=trava,
                                        n_evaluators=n_evaluators,
                                        test_result_handlers=test_result_handlers)

    tracker_mock = mocker.Mock()
    tracker_copy_func = mocker.patch.object(trava, '_tracker_copy')
    tracker_copy_func.return_value = tracker_mock

    scorer_1 = mocker.Mock()
    scorer_2 = mocker.Mock()

    scorers = [scorer_1, scorer_2]

    trava.evaluate_track(model_id=main_model_id if n_evaluators > 1 else model_id,
                         scorers=scorers)
    calls = [mocker.call.add_scorers(scorers=scorers)]
    for idx, evaluator in enumerate(evaluators):
        calls += [
            mocker.call.start_tracking(model_id=evaluator.model_id),
            mocker.call.track_model_results(model_results=mocker.ANY),
            mocker.call.end_tracking(model_id=evaluator.model_id),
        ]

        track_model_results_kwargs = tracker_mock.track_model_results.call_args_list[idx][1]
        model_results = track_model_results_kwargs['model_results']

        assert model_results.model_id == evaluator.model_id \
               and len(model_results.evaluators) == 1 \
               and model_results.evaluators == [evaluator]

    if n_evaluators > 1:
        calls += [
            mocker.call.start_tracking(model_id=main_model_id),
            mocker.call.track_model_results(model_results=mocker.ANY),
            mocker.call.end_tracking(model_id=main_model_id),
        ]

        track_model_results_kwargs = tracker_mock.track_model_results.call_args_list[n_evaluators][1]
        model_results = track_model_results_kwargs['model_results']

        assert model_results.model_id == main_model_id \
               and len(model_results.evaluators) == n_evaluators \
               and model_results.evaluators == evaluators

    tracker_mock.assert_has_calls(calls)


def test_detailed_results_no_model(trava):
    with pytest.raises(AttributeError):
        trava.detailed_results_for(model_id='arbitrary_id')


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_detailed_results(mocker,
                          trava,
                          model_id,
                          main_model_id,
                          model_type,
                          trava_model,
                          split_result,
                          test_result_handlers,
                          X_y_train,
                          X_y_test,
                          n_evaluators):
    test_model_id = main_model_id if n_evaluators > 1 else model_id

    meta_test_data = 1233
    trava_model.meta_test_data = meta_test_data

    result_handler = TestResultsHandler(scorers=[
        TestScorer(score_func=np.sum),
        TestAnyScorer(model_func=lambda scorer_model: scorer_model.meta_test_data)
    ])
    _fit_with_n_evaluators(mocker=mocker,
                           trava_model=trava_model,
                           model_id=model_id,
                           main_model_id=main_model_id,
                           model_type=model_type,
                           split_result=split_result,
                           trava=trava,
                           n_evaluators=n_evaluators,
                           test_result_handlers=[result_handler],
                           mock_evaluators=False)

    result = trava.detailed_results_for(model_id=test_model_id)

    X_train, _ = X_y_train
    X_test, _ = X_y_test

    f1_train_sum = np.sum(X_train['f1'])
    f1_test_sum = np.sum(X_test['f1'])

    assert result == [[[f1_train_sum] * n_evaluators, [f1_test_sum] * n_evaluators, [meta_test_data] * n_evaluators]]


# TODO: move to trava_base tests
@pytest.mark.parametrize("n_fits", [1, 2])
def test_results(mocker,
                 trava,
                 model_id,
                 main_model_id,
                 model_type,
                 trava_model,
                 split_result,
                 test_result_handlers,
                 X_y_train,
                 X_y_test,
                 n_fits):
    score_funcs = [
        np.sum,
        np.mean
    ]

    X_train, _ = X_y_train
    X_test, _ = X_y_test

    meta_test_data = 1233
    trava_model.meta_test_data = meta_test_data

    metrics_values = []
    for i in range(n_fits):
        score_func = score_funcs[i]
        result_handler = TestResultsHandler(scorers=[
            TestScorer(score_func=score_func),
            TestAnyScorer(model_func=lambda scorer_model: scorer_model.meta_test_data)
        ])
        _fit_with_n_evaluators(mocker=mocker,
                               trava_model=trava_model,
                               model_id=model_id + '_' + str(i),
                               main_model_id=main_model_id,
                               model_type=model_type,
                               split_result=split_result,
                               trava=trava,
                               n_evaluators=1,
                               test_result_handlers=[result_handler],
                               mock_evaluators=False)

        metrics_values.append((score_func(X_train['f1']), score_func(X_test['f1']), meta_test_data))

    results = trava.results

    test_metrics_values = [list(tup) for tup in list(zip(*metrics_values))]
    # test_metrics_values.append([])
    assert results == [test_metrics_values]


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_evaluate_dont_keep_data(mocker,
                                 trava,
                                 model_id,
                                 main_model_id,
                                 model_type,
                                 trava_model,
                                 split_result,
                                 test_result_handlers,
                                 n_evaluators):
    evaluators = _fit_with_n_evaluators(mocker=mocker,
                                        trava_model=trava_model,
                                        model_id=model_id,
                                        main_model_id=main_model_id,
                                        model_type=model_type,
                                        split_result=split_result,
                                        trava=trava,
                                        n_evaluators=n_evaluators,
                                        test_result_handlers=test_result_handlers,
                                        keep_data_in_memory=True)

    for evaluator in evaluators:
        evaluator.unload_data.assert_not_called()

    evaluators = _fit_with_n_evaluators(mocker=mocker,
                                        trava_model=trava_model,
                                        model_id=model_id,
                                        main_model_id=main_model_id,
                                        model_type=model_type,
                                        split_result=split_result,
                                        trava=trava,
                                        n_evaluators=n_evaluators,
                                        test_result_handlers=test_result_handlers,
                                        keep_data_in_memory=False)

    for evaluator in evaluators:
        evaluator.unload_data.assert_called()


@pytest.mark.parametrize("n_evaluators", [1, 3])
def test_raw_models(mocker,
                    trava,
                    model_id,
                    main_model_id,
                    model_type,
                    raw_model,
                    trava_model,
                    split_result,
                    test_result_handlers,
                    n_evaluators):
    evaluators = _fit_with_n_evaluators(mocker=mocker,
                                        trava_model=trava_model,
                                        model_id=model_id,
                                        main_model_id=main_model_id,
                                        model_type=model_type,
                                        split_result=split_result,
                                        trava=trava,
                                        n_evaluators=n_evaluators,
                                        test_result_handlers=test_result_handlers,
                                        keep_data_in_memory=False,
                                        mock_evaluators=False)

    target_model_id = model_id if n_evaluators == 1 else main_model_id
    raw_models = trava.raw_models_for(model_id=target_model_id)

    test_raw_models = dict([(evaluator.model_id, evaluator.trava_model.raw_model) for evaluator in evaluators])

    assert raw_models == test_raw_models


def test_create_raw_model(mocker, trava):
    model_type = TestModel
    def_1_value = 'custom_value'
    req_param_value = 111
    model_init_params = {'required_param_1': mocker.Mock(),
                         'required_param_2': req_param_value,
                         'def_1': def_1_value}
    model, result_init_params = trava._create_raw_model(model_type=model_type,
                                                        model_init_params=model_init_params)

    test_init_params = {
        'required_param_2': req_param_value,
        'def_1': def_1_value,
        'def_2': 'tada',
        'def_5': False
    }

    assert isinstance(model, model_type)
    assert result_init_params == test_init_params


def test_allowed_init_params_types():
    test_init_params_types = (
        numbers.Number,
        str,
        bool,
    )
    assert TravaSV._trackable_init_params_types() == test_init_params_types



from abc import abstractmethod
from typing import Optional

import matplotlib.pyplot as plt

from trava.tracker import Tracker


class CommonBoostingEvalLogic:
    """
    Some models support evaluation with early-stopping ( like gradient boosting ).
    This class was written with boosting in mind. Probably it can be generalized
    thus it won't be the base class anymore.
    """
    def __init__(self,
                 needs_plot: bool,
                 eval_metric: str = None,
                 early_stopping_rounds: Optional[int] = 10):
        """
        Parameters
        ----------
        needs_plot: bool
            Whether to plot the results of evaluation at the end.
        eval_metric: str
            What metric to use for evaluation
        early_stopping_rounds: int
            Number of rounds to wait before perform early stopping
        """
        self._needs_plot = needs_plot
        self._eval_metric = eval_metric
        self._early_stopping_rounds = early_stopping_rounds

    def setup_eval(self,
                   fit_params,
                   X_train,
                   y_train,
                   X_eval,
                   y_eval) -> dict:
        """
        You must make all the preparations to setup evaluation.

        Parameters
        ----------
        fit_params: dict
            Custom params to use when calling model's fit method
        X_train: pandas dataframe
            Train features for the model
        y_train: numpy array, pandas series
            Train target for the model
        X_eval: pandas dataframe
            Eval features for the model
        y_eval: numpy array, pandas series
            Eval target for the model

        Returns
        -------
        Prepared fit_params
        """
        result = fit_params.copy()

        if self._eval_metric:
            result['eval_metric'] = self._eval_metric

        result['eval_set'] = self._eval_set(X_train=X_train,
                                            y_train=y_train,
                                            X_eval=X_eval,
                                            y_eval=y_eval)

        if self._early_stopping_rounds:
            result['early_stopping_rounds'] = self._early_stopping_rounds

        return result

    def _eval_set(self, X_train, y_train, X_eval, y_eval):
        eval_set = [(X_eval, y_eval)]

        if self._user_train_in_eval:
            eval_set = [(X_train, y_train)] + eval_set

        return eval_set

    def train_set_results(self, model) -> Optional[list]:
        """
        Gets resulting metrics of evaluation for train set.

        Parameters
        ----------
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.

        Returns
        -------
        List containing evaluation metrics for train set.
        """
        eval_results = self._evals_results(model=model)
        key = self._train_metrics_key(model=model)

        if not key:
            return None

        return eval_results[key][self._get_metric_name(model=model)]

    def eval_set_results(self, model) -> list:
        """
        Gets resulting metrics of evaluation for eval set.

        Parameters
        ----------
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.

        Returns
        -------
        List containing evaluation metrics for eval set.
        """
        eval_results = self._evals_results(model=model)
        key = self._eval_metrics_key(model=model)

        return eval_results[key][self._get_metric_name(model=model)]

    def plot_if_needed(self, model_id: str, model, tracker: Tracker):
        """
        If plot was requested during initialization, learning curves
        will be plotted when evaluation has finished.

        Parameters
        ----------
        model_id: str
            Model unique identifier, will be used for saving metrics etc
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.
        tracker: Tracker
            Tracker for saving plots & eval results
        """
        if self._needs_plot:
            metric_to_plot = self._get_metric_name(model=model)

            train_results = self.train_set_results(model=model)
            eval_results = self.eval_set_results(model=model)

            epochs = len(eval_results)
            x_axis = range(0, epochs)

            fig, ax = plt.subplots()
            if train_results is not None:
                ax.plot(x_axis, train_results, label='Train')

            ax.plot(x_axis, eval_results, label='Test')
            ax.legend()
            plt.ylabel(metric_to_plot)
            plt.title('{} {}'.format('{} validation results'.format(model_id), metric_to_plot))
            plt.show()

            plot_filename = 'eval_plot' + '_' + model_id
            tracker.track_plot(model_id=model_id, fig=fig, filename=plot_filename)

    def track_eval_metrics(self, model_id: str, model, tracker: Tracker):
        """
        Tracks eval results

        Parameters
        ----------
        model_id: str
            Model unique identifier, will be used for saving metrics etc
        model: sklearn-style model
            Model that supports fit, predict and predict_proba methods.
        tracker: Tracker
            Tracker for saving plots & eval results
        """
        metric_to_plot = self._get_metric_name(model=model)

        train_results = self.train_set_results(model=model)
        eval_results = self.eval_set_results(model=model)

        tracker.track_tag(model_id=model_id, tag_key='eval_metric', tag_value=metric_to_plot)
        tracker.track_metric_value(model_id=model_id, name='early_stopping_rounds', value=self._early_stopping_rounds)
        tracker.track_metric_value(model_id=model_id, name='stopped_iteration', value=len(eval_results) - 1)
        tracker.track_metric_value(model_id=model_id, name='best_iteration', value=self._best_iteration(model=model))

        self._track_eval_results(model_id=model_id,
                                 eval_results=train_results,
                                 metric=metric_to_plot,
                                 train=True,
                                 tracker=tracker)

        self._track_eval_results(model_id=model_id,
                                 eval_results=eval_results,
                                 metric=metric_to_plot,
                                 train=False,
                                 tracker=tracker)

    @staticmethod
    def _track_eval_results(model_id: str,
                            eval_results: list,
                            metric: str,
                            train: bool,
                            tracker: Tracker):
        metric_name = ('train' if train else 'eval') + '_' + metric

        for idx, value in enumerate(eval_results):
            tracker.track_metric_value(model_id=model_id, name=metric_name, value=value, step=idx)

    @abstractmethod
    def _best_iteration(self, model) -> int:
        pass

    @abstractmethod
    def _evals_results(self, model) -> dict:
        pass

    @abstractmethod
    def _train_metrics_key(self, model) -> Optional[str]:
        pass

    @abstractmethod
    def _eval_metrics_key(self, model) -> str:
        pass

    @property
    def _user_train_in_eval(self) -> bool:
        return True

    def _get_metric_name(self, model) -> str:
        eval_results = self._evals_results(model=model)
        result = list(eval_results[self._eval_metrics_key(model=model)].keys())[0]
        return result

    def _n_eval_sets(self, model) -> int:
        eval_results = self._evals_results(model=model)

        if not isinstance(eval_results, dict):
            raise ValueError('Eval results must be dict')

        n_eval_sets = len(eval_results.keys())
        assert 0 < n_eval_sets <= 2, "only train and test are now supported"

        return n_eval_sets



import pytest

from trava.fit_predictor import FitPredictConfig


@pytest.fixture(scope='class')
def model_id():
    return 'test_model_id'


@pytest.fixture()
def model(mocker, model_id):
    result = mocker.Mock()
    result.model_id = model_id
    return result


def test_equality(mocker, model, model_id):
    model_init_params = {'a': 1, 'b': 2}
    scorers_providers = mocker.MagicMock()
    serializer = mocker.Mock()
    split_result = mocker.Mock()
    raw_dataset = mocker.Mock()
    fit_params = {'c': 3, 'd': 4}
    predict_params = {'_': 11, '+': 43}
    description = 'descr'

    first_config = FitPredictConfig(raw_model=model,
                                    model_init_params=model_init_params,
                                    model_id=model_id,
                                    scorers_providers=scorers_providers,
                                    serializer=serializer,
                                    raw_split_data=split_result,
                                    raw_dataset=raw_dataset,
                                    fit_params=fit_params,
                                    predict_params=predict_params,
                                    description=description)

    second_config = FitPredictConfig(raw_model=model,
                                     model_init_params=model_init_params,
                                     model_id=model_id,
                                     scorers_providers=scorers_providers,
                                     serializer=serializer,
                                     raw_split_data=split_result,
                                     raw_dataset=raw_dataset,
                                     fit_params=fit_params,
                                     predict_params=predict_params,
                                     description=description)

    assert first_config == second_config
    second_config.serializer = mocker.Mock()
    assert first_config != second_config
    assert first_config != 'str'

    none_config = FitPredictConfig(raw_model=None,
                                   model_init_params=None,
                                   model_id='123',
                                   scorers_providers=[],
                                   serializer=None,
                                   raw_split_data=mocker.Mock())
    assert first_config != none_config



import inspect
import numbers
from abc import ABC
from typing import Optional, List

from trava.logger import TravaLogger
from trava.model_results import ModelResult
from trava.results_handler import ResultsHandler
from trava.trava_tracker import Tracker, TravaTracker


class _TravaBase(ABC):
    """
    Trava stands for TrainValidation.

    It will help you to gather metrics from different models you try
    and present them in any way that is convenient for you.

    Init parameters
    ----------
    logger: TravaLogger
        Encapsulates logic of work with native python logger
    tracker: Tracker
        Is used to track your experiments
    results_handlers: list
        Each results handler provides logic how to present the metrics to a user.
    """
    def __init__(self,
                 logger: Optional[TravaLogger] = None,
                 tracker: Optional[Tracker] = None,
                 results_handlers: List[ResultsHandler] = None):
        self._logger = logger or TravaLogger()
        self._tracker = tracker or TravaTracker(scorers=[])
        self._results_handlers = results_handlers

        if results_handlers is None:
            self._results_handlers = []

        self._results = {}

    @property
    def results(self) -> list:
        """
        Calls every results handler to present all the metrics
        for all the fits that were made so far.

        Returns
        -------
        List of outputs for every results handler that returns not None result
        """
        return self._results_for(results=list(self._results.values()), results_handlers=self._results_handlers)

    def _results_for(self, results: List[ModelResult], results_handlers: List[ResultsHandler]) -> list:
        """
        Applies all the handlers and returns their result

        Returns
        -------
        List of outputs for every results handler that returns not None result
        """
        result = []
        for handler in results_handlers:
            handler_result = handler.handle(results=results, logger=self._logger, tracker=self._tracker)
            if handler_result is not None:
                result.append(handler_result)

        return result

    def raw_models_for(self, model_id: str) -> Optional[list]:
        """
        Get raw models associated with the given model_id

        Parameters
        ----------
        model_id: str
            Unique identifier of the requested model

        Returns
        -------
        List of trained models if exists
        """
        model_results = self._model_results(model_id=model_id)
        return model_results.raw_models

    def _model_results(self, model_id: str) -> ModelResult:
        result = self._results.get(model_id)
        assert result, "You haven't fit a model with such model_id: {}".format(model_id)
        return result

    def _assert_existing_model(self, model_id: str):
        assert self._results.get(model_id) is None, "you tried to save a model with the already existing id"

    def _create_raw_model(self, model_type, model_init_params: dict) -> tuple:
        """
        Creates model object with the provided params.

        model_type: type of sklearn-style model
            Type of model that supports fit, predict and predict_proba methods
        model_init_params: dict
            Parameters to use to initialize model_type
        :return:
        Initialized model + trackable init parameters, including default ones.
        See _trackable_init_params_types for supported param types.
        """
        raw_model = model_type(**model_init_params)
        all_init_params = self._merge_given_params_with_default(model_type=model_type,
                                                                model_init_params=model_init_params)

        return raw_model, all_init_params

    @staticmethod
    def _merge_given_params_with_default(model_type, model_init_params: dict) -> dict:
        """
        Merges default params for model_type with params given by a user
        """
        all_params = dict(model_init_params)
        for key, param in inspect.signature(model_type.__init__).parameters.items():
            if not model_init_params.get(key) and param.default != param.empty:
                all_params[key] = param.default

        result = {}
        for key, value in all_params.items():
            if isinstance(value, _TravaBase._trackable_init_params_types()):
                result[key] = value

        return result

    @staticmethod
    def _trackable_init_params_types() -> tuple:
        """
        Types for a model's init params that Trava is able to track.
        """
        return (
            numbers.Number,
            str,
            bool,
        )



from sklearn.metrics import make_scorer

from trava.model_info import ModelInfo
from trava.scorer import Scorer


class SklearnScorer(Scorer):
    """
    Uses standard sklearn's make_scorer.
    """
    def __init__(self,
                 score_func: callable,
                 greater_is_better=True,
                 needs_proba=False,
                 needs_threshold=False,
                 sample_weight_required=False,
                 **metrics_kwargs):
        self._greater_is_better = greater_is_better
        self._needs_threshold = needs_threshold
        self._sample_weight_required = sample_weight_required

        super().__init__(score_func=score_func,
                         needs_proba=needs_proba,
                         requires_raw_model=needs_threshold,
                         requires_X_y=False,
                         **metrics_kwargs)

    def _make_scorer(self, score_func: callable, **metrics_kwargs) -> callable:
        def scorer(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
            if self._sample_weight_required and X is None:
                raise Exception('Sample weight is required for score ({}) calculation, '
                                'so it must be called with valid X_y data'.format(self.func_name))

            if X is None:
                # the last resort. If data and model are unloaded and your metric support interface
                # like (y_true, y_pred, **kwargs), then you will be able to calculate metrics

                y_cached = model_info.y(for_train=for_train)

                if self._needs_proba:
                    y_pred_values = model_info.y_pred_proba(for_train=for_train)
                    if len(set(y_cached)) == 2:
                        y_pred_values = y_pred_values[:, 1]
                else:
                    y_pred_values = model_info.y_pred(for_train=for_train)

                result = score_func(y_cached, y_pred_values, **metrics_kwargs)
            else:
                sk_scorer = self._get_sklearn_scorer(score_func=score_func,
                                                     **metrics_kwargs)
                sample_weight = self._sample_weight(X=X, X_raw=X_raw)
                result = sk_scorer(estimator=model,
                                   X=X,
                                   y_true=y,
                                   sample_weight=sample_weight)
            return result
        return scorer

    def _get_sklearn_scorer(self, score_func: callable, **metrics_kwargs):
        result = make_scorer(score_func=score_func,
                             greater_is_better=self._greater_is_better,
                             needs_proba=self._needs_proba,
                             needs_threshold=self._needs_threshold,
                             **metrics_kwargs)
        return result

    @staticmethod
    def _sample_weight(X, X_raw):
        """
        Maybe it will help to handle cases when the parameter is needed in a scorer.
        Introduced this template method just to not forget about it.
        """
        pass


# wrappers for sklearn metrics

def sk(score_func: callable, **kwargs) -> Scorer:
    return _sk(score_func=score_func, needs_proba=False, **kwargs)


def sk_proba(score_func: callable, **kwargs) -> Scorer:
    return _sk(score_func=score_func, needs_proba=True, **kwargs)


def _sk(score_func: callable, needs_proba=False,  **kwargs) -> Scorer:
    return SklearnScorer(score_func=score_func, needs_proba=needs_proba, **kwargs)



import pytest
import numpy as np

from trava.trava_model import TravaModel, _CachedModel


@pytest.fixture(scope='class')
def model_id():
    return 'test_model_id'


@pytest.fixture
def raw_model(mocker):
    return mocker.Mock()


@pytest.fixture
def X(mocker):
    return mocker.Mock()


@pytest.fixture
def y(mocker):
    return mocker.Mock()


@pytest.fixture
def fit_params(mocker):
    return {'fit_param': 111}


@pytest.fixture
def predict_params(mocker):
    return {'predict_param': 222}


def test_copy(mocker, model_id):
    raw_model = mocker.Mock()
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    y_train = np.array([0, 0, 1])
    y_train_pred = np.array([1, 2, 3])
    y_train_pred_proba = np.array([3, 4, 5])
    y_test = np.array([-2, 3, 5])
    y_test_pred = np.array([6, 7, 8])
    y_test_pred_proba = np.array([9, 10, 11])

    fit_params = {'1': 2}
    predict_params = {'2': 3}
    fit_time = 123
    predict_time = 434

    model._y_train = y_train
    model._y_train_pred = y_train_pred
    model._y_train_pred_proba = y_train_pred_proba

    model._y_test = y_test
    model._y_test_pred = y_test_pred
    model._y_test_pred_proba = y_test_pred_proba

    model._fit_params = fit_params
    model._predict_params = predict_params
    model._fit_time = fit_time
    model._predict_time = predict_time

    model_copy_id = model_id + '_copy'
    model_copy = model.copy(model_id=model_copy_id)

    assert model_copy.model_id == model_copy_id
    y = model.y(for_train=True)
    copy_y = model_copy.y(for_train=True)
    assert np.array_equal(y, copy_y)
    assert np.array_equal(model.y(for_train=False), model_copy.y(for_train=False))
    assert np.array_equal(model.y_pred(for_train=False), model_copy.y_pred(for_train=False))
    assert np.array_equal(model.y_pred_proba(for_train=False), model_copy.y_pred_proba(for_train=False))
    assert model.fit_params == model_copy.fit_params
    assert model.predict_params == model_copy.predict_params
    assert model.fit_time == model_copy.fit_time
    assert model.predict_time == model_copy.predict_time


def test_model_id(mocker, model_id):
    raw_model = mocker.Mock()
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    assert model.model_id == model_id


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_get_model(mocker, raw_model, X, y, needs_proba):
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    assert model.get_model(for_train=True) == raw_model
    assert model.get_model(for_train=False) == raw_model

    y_predict_proba = mocker.Mock()
    if needs_proba:
        raw_model.predict_proba.return_value = y_predict_proba

    y_pred = mocker.Mock()
    raw_model.predict.return_value = y_pred

    model.fit(X=X, y=y)
    model.predict(X=X, y=y)

    model.unload_model()

    train_cached_model = model.get_model(for_train=True)
    test_cached_model = model.get_model(for_train=False)

    assert train_cached_model != raw_model
    assert test_cached_model != raw_model

    assert train_cached_model.predict(X) == y_pred
    if needs_proba:
        assert train_cached_model.predict_proba(X) == y_predict_proba


@pytest.mark.parametrize("for_train", [True, False], ids=["for_train", "for_test"])
def test_get_model_unload(mocker, raw_model, for_train):
    trava_model = TravaModel(raw_model=raw_model, model_id=model_id)
    trava_model.unload_model()

    with pytest.raises(ValueError):
        trava_model.get_model(for_train=for_train)

    if for_train:
        y_pred_key = '_y_train_pred'
    else:
        y_pred_key = '_y_test_pred'

    y_pred_mock = mocker.Mock()
    mocker.patch.object(trava_model, y_pred_key, y_pred_mock)

    assert trava_model.get_model(for_train=for_train).predict(X=None) == y_pred_mock


def test_raw_model(mocker, raw_model, model_id):
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    assert model.raw_model == raw_model


def test_fit_params(mocker, raw_model, model_id, X, y, fit_params, predict_params):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    assert model.fit_params == fit_params


def test_predict_params(mocker, raw_model, model_id, X, y, fit_params, predict_params):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    assert model.predict_params == predict_params


def test_fit_time(mocker, raw_model, model_id, X, y):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    assert not model.fit_time
    model.fit(X=X, y=y)
    assert model.fit_time


def test_predict_time(mocker, raw_model, model_id, X, y):
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    assert not model.predict_time
    model.predict(X=X, y=y)
    assert model.predict_time


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_fit(mocker, raw_model, model_id, X, y, fit_params, predict_params, needs_proba):
    if needs_proba:
        predict_proba = mocker.Mock()
        raw_model.predict_proba.return_value = predict_proba

    y_pred = mocker.Mock()
    raw_model.predict.return_value = y_pred

    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    raw_model.fit.assert_called_once_with(X, y, **fit_params)
    raw_model.predict.assert_called_once_with(X, **predict_params)

    if needs_proba:
        raw_model.predict_proba.assert_called_with(X, **predict_params)

    assert model.fit_time


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_predict(mocker, raw_model, model_id, X, y, fit_params, needs_proba):
    if needs_proba:
        predict_proba = mocker.Mock()
        raw_model.predict_proba.return_value = predict_proba

    y_pred = mocker.Mock()
    raw_model.predict.return_value = y_pred

    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.predict(X=X, y=y)

    raw_model.predict.assert_called_once_with(X)

    if needs_proba:
        raw_model.predict_proba.assert_called_with(X)

    assert model.predict_time


@pytest.mark.parametrize("needs_proba", [True, False], ids=["proba", "no_proba"])
def test_all_y(mocker, raw_model, model_id, X, y, fit_params, predict_params, needs_proba):
    model = TravaModel(raw_model=raw_model, model_id=model_id)

    predict_proba_train = mocker.Mock()
    if needs_proba:
        raw_model.predict_proba.return_value = predict_proba_train
    y_pred_train = mocker.Mock()
    raw_model.predict.return_value = y_pred_train
    model.fit(X=X, y=y, fit_params=fit_params, predict_params=predict_params)

    predict_proba_test = mocker.Mock()
    if needs_proba:
        raw_model.predict_proba.return_value = predict_proba_test
    y_pred_test = mocker.Mock()
    raw_model.predict.return_value = y_pred_test
    X_test = mocker.Mock()
    y_test = mocker.Mock()
    model.predict(X=X_test, y=y_test)

    assert model.y_pred(for_train=True) == y_pred_train
    assert model.y_pred(for_train=False) == y_pred_test
    assert model.y(for_train=True) == y
    assert model.y(for_train=False) == y_test
    if needs_proba:
        assert model.y_pred_proba(for_train=True) == predict_proba_train
        assert model.y_pred_proba(for_train=False) == predict_proba_test


def test_unload(mocker, model_id):
    raw_model = mocker.Mock()
    model = TravaModel(raw_model=raw_model, model_id=model_id)
    model.unload_model()

    assert not model.raw_model



from typing import List

from trava.split.config import DataSplitConfig
from trava.split.result import SplitResult


# noinspection PyPep8Naming
class Splitter:
    @staticmethod
    def split(df,
              config: DataSplitConfig,
              **kwargs):
        split_logic = config.split_logic

        train_df, test_df, valid_df = split_logic.split(data=df,
                                                        test_size=config.test_size,
                                                        valid_size=config.valid_size,
                                                        **kwargs)

        def split_X_y(source_df):
            return Splitter._split_X_y(df=source_df,
                                       ignore_cols=config.ignore_cols,
                                       target_col_name=config.target_col_name)

        # TODO: change cols to indices
        X_train, y_train = split_X_y(source_df=train_df)

        X_valid, y_eval = None, None
        if valid_df is not None:
            X_valid, y_eval = split_X_y(source_df=valid_df)

        X_test, y_test = split_X_y(source_df=test_df)

        result = SplitResult(X_train=X_train,
                             X_test=X_test,
                             y_train=y_train,
                             y_test=y_test,
                             X_valid=X_valid,
                             y_valid=y_eval)

        for handler in config.split_result_handlers:
            result = handler.handle(split_result=result)

        return result

    @staticmethod
    def _split_X_y(df,
                   ignore_cols: List[str],
                   target_col_name: str):
        X, y = df.drop(ignore_cols + [target_col_name], axis=1), df[target_col_name]
        return X, y



from typing import Optional
from sklearn.model_selection import train_test_split

from trava.split.logic import SplitLogic


class BasicSplitLogic(SplitLogic):
    """
    Basically just a wrapper around sklearn's train_test_split.
    Can be used in combination with another split logic.
    """
    def __init__(self,
                 random_state: int = 15,
                 shuffle: bool = True,
                 eval_from_test: bool = False,
                 group_col: Optional[str] = None,
                 stratify: bool = None):
        self._random_state = random_state
        self._shuffle = shuffle
        self._stratify = stratify
        self._valid_from_test = eval_from_test
        self._group_col = group_col

    def split(self,
              data,
              test_size: float,
              valid_size: float = 0.0,
              **kwargs) -> tuple:
        data_to_split = data

        if self._group_col:
            unique_groups = data[self._group_col].unique()
            data_to_split = unique_groups

        train_data, test_data = train_test_split(data_to_split,
                                                 test_size=test_size,
                                                 shuffle=self._shuffle,
                                                 stratify=self._stratify,
                                                 random_state=self._random_state,
                                                 **kwargs)

        valid_data = None
        if valid_size > 0.0:
            data_for_valid_split = test_data if self._valid_from_test else data_to_split
            test_data, valid_data = train_test_split(data_for_valid_split,
                                                     test_size=valid_size,
                                                     shuffle=self._shuffle,
                                                     stratify=self._stratify,
                                                     random_state=self._random_state,
                                                     **kwargs)

        if self._group_col:
            train_data = data[data[self._group_col].isin(train_data)]
            test_data = data[data[self._group_col].isin(test_data)]
            if valid_data is not None:
                valid_data = data[data[self._group_col].isin(valid_data)]

        return train_data, test_data, valid_data



from typing import Optional

import pytest

from tests.objects_for_tests import TestScorer, TestAnyScorer
from trava.model_info import ModelInfo
from trava.scorer import Scorer


@pytest.fixture
def score_func(mocker, score_func_name):
    result = mocker.Mock()
    result.__name__ = score_func_name

    return result


@pytest.fixture(scope='class')
def score_func_name():
    return 'test_score_func'


def test_is_any_scorer(mocker, score_func):
    scorer = TestScorer(score_func=score_func)
    assert not scorer.is_other_scorer
    any_scorer = TestAnyScorer(model_func=score_func)
    assert any_scorer.is_other_scorer


def test_func_name(mocker, score_func, score_func_name):
    scorer = TestScorer(score_func=score_func)
    assert scorer._func_name == score_func_name


def test_require_raw_model(mocker, score_func):
    scorer = TestScorer(score_func=score_func, requires_raw_model=True)
    trava_model = mocker.Mock()
    trava_model.raw_model.return_value = None

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()

    with pytest.raises(Exception):
        scorer(trava_model=trava_model,
               for_train=True,
               X=X,
               X_raw=X_raw,
               y=y)

    trava_model.raw_model.return_value = mocker.MagicMock()

    scorer(trava_model=trava_model,
           for_train=True,
           X=X,
           X_raw=X_raw,
           y=y)


def test_require_X_y(mocker, score_func):
    scorer = TestScorer(score_func=score_func, requires_X_y=True)
    trava_model = mocker.Mock()
    trava_model.raw_model.return_value = mocker.MagicMock()

    X = None
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()

    with pytest.raises(Exception):
        scorer(trava_model=trava_model,
               for_train=True,
               X=X,
               X_raw=X_raw,
               y=y)

    X = mocker.MagicMock()
    X_raw = None
    y = mocker.MagicMock()

    with pytest.raises(Exception):
        scorer(trava_model=trava_model,
               for_train=True,
               X=X,
               X_raw=X_raw,
               y=y)

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = None

    with pytest.raises(Exception):
        scorer(trava_model=trava_model,
               for_train=True,
               X=X,
               X_raw=X_raw,
               y=y)

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()
    scorer(trava_model=trava_model,
           for_train=True,
           X=X,
           X_raw=X_raw,
           y=y)


@pytest.mark.parametrize("for_train", [True, False])
def test_scorer_func_input(mocker, score_func, for_train):
    kwargs = {'a': 3, 'b': '3fff'}
    scorer = TestScorer(score_func=score_func, **kwargs)
    internal_scorer = mocker.patch.object(scorer, '_scorer')

    trava_model = mocker.Mock()
    trava_model.raw_model.return_value = mocker.MagicMock()
    raw_model = mocker.Mock()
    trava_model.get_model.return_value = raw_model

    X = mocker.MagicMock()
    X_raw = mocker.MagicMock()
    y = mocker.MagicMock()

    scorer(trava_model=trava_model,
           for_train=for_train,
           X=X,
           X_raw=X_raw,
           y=y,
           **kwargs)

    internal_scorer.assert_called_once_with(model=raw_model,
                                            model_info=trava_model,
                                            for_train=for_train,
                                            X=X,
                                            X_raw=X_raw,
                                            y=y,
                                            **kwargs)


def test_make_scorer(mocker, score_func):
    class LocalTestScorer(Scorer):
        def __init__(self, score_func: callable, needs_proba=False, requires_raw_model=False,
                     requires_X_y=False, name: Optional[str] = None, **metrics_kwargs):
            self._make_scorer_score_func = None
            self._make_scorer_metrics_kwargs = None
            super().__init__(score_func, needs_proba, requires_raw_model, requires_X_y, name, **metrics_kwargs)

        def _make_scorer(self, score_func: callable, **metrics_kwargs) -> callable:
            self._make_scorer_score_func = score_func
            self._make_scorer_metrics_kwargs = metrics_kwargs

            def scorer(model, model_info: ModelInfo, for_train: bool, X, X_raw, y):
                return 1

            return scorer

    kwargs = {'a': 3, 'b': '3fff'}
    scorer = LocalTestScorer(score_func=score_func, **kwargs)

    assert scorer._make_scorer_score_func == score_func
    assert scorer._make_scorer_metrics_kwargs == kwargs


